#include <stdio.h>
#include "Rinternals.h"
#include "R_ext/Rdynload.h"
static SEXP WFILE_open(SEXP name, SEXP mode);
static SEXP WFILE_close(SEXP s);
static SEXP WFILE_gets(SEXP s);
extern void R_PreserveObject(SEXP);
static void AddFileRef(SEXP ref);
static SEXP WFILE_list(void);
static SEXP WFILE_type_tag;
#define CHECK_WFILE_STREAM(s) do { \
    if (TYPEOF(s) != EXTPTRSXP || \
        R_ExternalPtrTag(s) != WFILE_type_tag) \
        error("bad file stream"); \
} while (0)
static SEXP FileList;
static R_CallMethodDef WFILE_CallDefs[] = {
    {"WFILE_open", (DL_FUNC) WFILE_open, 2},
    {"WFILE_close", (DL_FUNC) WFILE_close, 1},
    {"WFILE_gets", (DL_FUNC) WFILE_gets, 1},
    {"WFILE_list", (DL_FUNC) WFILE_list, 0},
    {NULL}
};
static SEXP WFILE_open(SEXP name, SEXP mode)
{
    FILE *f = fopen(CHAR(STRING_ELT(name, 0)), CHAR(STRING_ELT(mode, 0)));
    if (f == NULL)
        return R_NilValue;
    else {
        SEXP val, ref;
        PROTECT(val = R_MakeExternalPtr(f, WFILE_type_tag, R_NilValue));
        PROTECT(ref = R_MakeWeakRefC(val, name,
                                     (R_CFinalizer_t) WFILE_close, TRUE));
        AddFileRef(ref);
        UNPROTECT(2);
        return val;
    }
}
static SEXP WFILE_close(SEXP s)
{
    FILE *f;
    CHECK_WFILE_STREAM(s);
    f = R_ExternalPtrAddr(s);
    if (f != NULL) {
        fclose(f);
        R_ClearExternalPtr(s);
    }
    return R_NilValue;
}
static SEXP WFILE_gets(SEXP s)
{
    char buf[512];
    FILE *f;
    CHECK_WFILE_STREAM(s);
    f = R_ExternalPtrAddr(s);
    if (f == NULL)
        error("file pointer is NULL");
    if (fgets(buf, sizeof(buf), f) == NULL)
        return R_NilValue;
    else {
        SEXP val;
        PROTECT(val = allocVector(STRSXP, 1));
        SET_STRING_ELT(val, 0, mkChar(buf));
        UNPROTECT(1);
        return val;
    }
}
static void AddFileRef(SEXP ref)
{
    SEXP f, files, next = NULL, last = NULL;
    files = CDR(FileList);
    for (f = files; f != R_NilValue; f = next) {
        SEXP ref = CAR(f);
        SEXP key = R_WeakRefKey(ref);
        next = CDR(f);
        if (key == R_NilValue ||  R_ExternalPtrAddr(key) == NULL) {
            if (last == NULL) files = next;
            else SETCDR(last, next);
        }
        else last = f;
    }
    SETCDR(FileList, CONS(ref, files));
}
static SEXP WFILE_list(void)
{
    SEXP files, val = R_NilValue;
    for (files = CDR(FileList); files != R_NilValue; files = CDR(files)) {
        SEXP ref = CAR(files);
        SEXP key = R_WeakRefKey(ref);
        if (key != R_NilValue && R_ExternalPtrAddr(key) != NULL) {
            PROTECT(key);
            val = CONS(R_WeakRefValue(ref), val);
            UNPROTECT(1);
        }
    }
    return PairToVectorList(val);
}
void R_init_wfile(DllInfo *info)
{
    WFILE_type_tag = install("WFILE_TYPE_TAG");
    FileList = CONS(R_NilValue, R_NilValue);
    R_PreserveObject(FileList);
    R_registerRoutines(info, NULL, WFILE_CallDefs, NULL, 0);
}
